from contextvars import ContextVar

from giving import give

current_overseer = ContextVar("current_overseer", default=None)


def log(**kwargs):
    ov = current_overseer.get()
    if ov is not None:
        ov.log(kwargs)


def iterate(task: str, iterable, report_batch=False):
    assert isinstance(task, str)
    n = len(iterable)
    with give.inherit(task=task):
        give(progress=(0, n))
        for i, batch in enumerate(iterable):
            if report_batch:
                with give.wrap("step", batch=batch):
                    yield batch
            else:
                yield batch
            give(progress=(i + 1, n))



import time
from typing import Union

from giving import Given
from ovld import ovld

from ..tools import instrument_definition


@ovld
def _parse_duration(x: Union[int, float]):
    return (False, x)


@ovld
def _parse_duration(x: str):  # noqa: F811
    if x.endswith("s"):
        return (True, float(x[:-1]))
    else:
        return (False, float(x))


@instrument_definition
def rate(ov, interval=1, method=None, multimodal_batch=True, sync=None):

    if multimodal_batch:

        def _batchsize(batch):
            if isinstance(batch, (list, tuple)):
                return len(batch[0])
            else:
                return len(batch)

    else:
        _batchsize = len

    yield ov.phases.load_script

    interval_is_time, interval = _parse_duration(interval)

    # Build stream of task/time/batch_size
    if method is None or method == "step":
        steps_w_batch = ov.given.where("task", "batch", "!batch_size", "!$wrap").kmap(
            task=lambda task: task, batch_size=lambda batch: _batchsize(batch)
        )
        steps_w_batch_size = ov.given.where("task", "batch_size", "!$wrap")
        times = times_step = (
            (steps_w_batch | steps_w_batch_size)
            .augment(time=lambda: time.time_ns())
            .pairwise()
            .starmap(
                lambda x, y: {
                    "task": y["task"],
                    "time": (y["time"] - x["time"]) / 1_000_000_000,
                    "batch_size": y["batch_size"],
                }
            )
        )

    if method is None or method == "wrap":

        def _timewrap():
            t0 = time.time_ns()
            results = yield
            t1 = time.time_ns()
            task = results["task"]
            if "batch_size" in results:
                seconds = (t1 - t0) / 1_000_000_000
                return {
                    "task": task,
                    "time": seconds,
                    "batch_size": results["batch_size"],
                }
            elif "batch" in results:
                seconds = (t1 - t0) / 1_000_000_000
                data = results["batch"]
                return {"task": task, "time": seconds, "batch_size": _batchsize(data)}
            else:
                return None

        times = times_wrap = ov.given.wmap("step", _timewrap).filter(lambda x: x)

    if method is None:
        times = times_step | times_wrap

    grouped_by_task = times.group_by(lambda data: data["task"])

    @grouped_by_task.subscribe
    def setup_pipeline(times):
        times = Given(_obs=times)

        # Group by interval
        if interval_is_time:
            times = times.buffer_with_time(interval)
        else:
            times = times.buffer_with_count(interval)

        # Compute the final metric
        @times.subscribe
        def _(elems):
            t = 0
            if sync is not None:
                t0 = time.time_ns()
                sync()
                t1 = time.time_ns()
                t += (t1 - t0) / 1_000_000_000

            t += sum(e["time"] for e in elems)
            n = sum(e["batch_size"] for e in elems)

            if n and t:
                ov.give(rate=n / t, units="items/s", task=elems[0]["task"])



import json
import os
import shutil
import subprocess
import sys
import time
from threading import Thread

from ..tools import instrument_definition

nvml_available = None


def nvlm_init():
    global nvml_available

    if nvml_available is None:
        from pynvml import nvmlInit
        from pynvml.nvml import (
            NVMLError_DriverNotLoaded,
            NVMLError_LibraryNotFound,
        )

        nvml_available = False
        try:
            nvmlInit()
            nvml_available = True
        except NVMLError_LibraryNotFound:
            pass
        except NVMLError_DriverNotLoaded:
            pass

    return nvml_available


def get_cuda_info():
    from pynvml.smi import nvidia_smi

    nvml_available = nvlm_init()

    def fix_num(n):
        if n == "N/A":
            n = None
        return n

    def parse_gpu(gpu, gid):
        mem = gpu["fb_memory_usage"]
        used = fix_num(mem["used"])
        total = fix_num(mem["total"])
        compute = fix_num(gpu["utilization"]["gpu_util"])
        if compute:
            compute /= 100
        return {
            "device": gid,
            "product": gpu["product_name"],
            "memory": {
                "used": used,
                "total": total,
            },
            "utilization": {
                "compute": compute,
                "memory": total and used and (used / total),
            },
            "temperature": fix_num(gpu["temperature"]["gpu_temp"]),
            "power": fix_num(gpu["power_readings"]["power_draw"]),
            "selection_variable": "CUDA_VISIBLE_DEVICES",
        }

    if not nvml_available:
        return {}

    nvsmi = nvidia_smi.getInstance()

    to_query = [
        "gpu_name",
        "memory.free",
        "memory.used",
        "memory.total",
        "temperature.gpu",
        "utilization.gpu",
        "utilization.memory",
        "power.draw",
    ]
    results = nvsmi.DeviceQuery(",".join(to_query))

    if not results or "gpu" not in results:
        return {}

    gpus = results["gpu"]
    if not isinstance(gpus, list):
        gpus = [gpus]

    return {str(i): parse_gpu(g, i) for i, g in enumerate(gpus)}


def _get_info(requires, command, parse_function):
    if not shutil.which(requires):
        return None
    proc_results = subprocess.run(
        command, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    try:
        return parse_function(json.loads(proc_results.stdout))
    except json.JSONDecodeError:
        print(f"There was a problem with {requires}:")
        print("=" * 80)
        print(proc_results.stderr, file=sys.stderr)
        print("=" * 80)
        return None


def get_rocm_info():
    return _get_info("rocm-smi", "rocm-smi -a --showmeminfo vram --json", parse_rocm)


def parse_rocm(info):
    def parse_gpu(gpu, gid):
        used = int(gpu["VRAM Total Used Memory (B)"])
        total = int(gpu["VRAM Total Memory (B)"])
        return {
            "device": gid,
            "product": "ROCm Device",
            "memory": {
                "used": used // (1024**2),
                "total": total // (1024**2),
            },
            "utilization": {
                "compute": float(gpu["GPU use (%)"]) / 100,
                "memory": used / total,
            },
            "temperature": float(gpu["Temperature (Sensor edge) (C)"]),
            "power": float(gpu["Average Graphics Package Power (W)"]),
            "selection_variable": "ROCR_VISIBLE_DEVICES",
        }

    results = {}
    for k, v in info.items():
        x, y, cnum = k.partition("card")
        if x != "" or y != "card":
            continue
        results[cnum] = parse_gpu(v, cnum)

    return results


def get_gpu_info(arch=None):
    if arch is None:
        cuda = get_cuda_info()
        rocm = get_rocm_info()
        if cuda and rocm:
            raise Exception(
                "Milabench found both CUDA and ROCM-compatible GPUs and does not"
                " know which kind to use. Please set $MILABENCH_GPU_ARCH to 'cuda',"
                " 'rocm' or 'cpu'."
            )
        elif cuda:
            arch = "cuda"
            results = cuda
        elif rocm:
            arch = "rocm"
            results = rocm
        else:
            arch = "cpu"
            results = {}

    elif arch == "cuda":
        results = get_cuda_info()
    elif arch == "rocm":
        results = get_rocm_info()
    elif arch == "cpu":
        results = {}
    else:
        raise ValueError(
            "Could not infer the gpu architecture because both cuda "
            "and rocm-compatible gpus were found. Please specify "
            "arch in ('cuda', 'rocm', 'cpu')"
        )

    return {"arch": arch, "gpus": results}


class Monitor(Thread):
    def __init__(self, ov, delay, func):
        super().__init__(daemon=True)
        self.ov = ov
        self.stopped = False
        self.delay = delay
        self.func = func

    def run(self):
        while not self.stopped:
            time.sleep(self.delay)
            self.func()

    def stop(self):
        self.stopped = True


@instrument_definition
def gpu_monitor(ov, poll_interval=10, arch=None):
    yield ov.phases.load_script

    visible = os.environ.get("CUDA_VISIBLE_DEVICES", None) or os.environ.get(
        "ROCR_VISIBLE_DEVICES", None
    )
    if visible:
        ours = visible.split(",")
    else:
        ours = [str(x) for x in range(100)]

    def monitor():
        data = {
            gpu["device"]: {
                "memory": [gpu["memory"]["used"], gpu["memory"]["total"]],
                "load": gpu["utilization"]["compute"],
                "temperature": gpu["temperature"],
            }
            for gpu in get_gpu_info(arch)["gpus"].values()
            if str(gpu["device"]) in ours
        }
        ov.give(task="main", gpudata=data)

    monitor_thread = Monitor(ov, poll_interval, monitor)
    monitor_thread.start()
    try:
        yield ov.phases.run_script
    finally:
        monitor_thread.stop()
        monitor()
